"""This module is meant to be used only if attrs library is installed. The library is an opt-in dependency."""

from typing import Any, Optional, Tuple

from attrs import NOTHING, Factory, fields, has

from ...docstring import DocstringParser
from ..types.dict import ConceptDictionaryType, ConceptDictionaryTypeField


def is_attrs(annotation: Any) -> bool:
    return has(annotation)


def compute_default_value(field: Any) -> Tuple[bool, Any]:
    default_value = field.default
    if default_value is not NOTHING:
        if isinstance(default_value, Factory):  # type: ignore[reportArgumentType]
            return True, default_value.factory()

        return True, default_value

    return False, None


def from_attrs(annotation, factory, unset) -> ConceptDictionaryType:
    docstring = DocstringParser.parse(annotation.__doc__)
    unset = getattr(annotation, "__unset__", unset)

    dict_fields = []

    for field in fields(annotation):
        description: Optional[str] = next((attribute.description for attribute in docstring.attributes if attribute.name == field.name), None)

        has_default_value, default_value = compute_default_value(field)

        if field.init is False:
            concept_type = factory.from_type(Optional[field.type], unset=unset)  # type: ignore
        else:
            concept_type = factory.from_type(field.type, unset=unset)  # type: ignore

        dict_fields.append(ConceptDictionaryTypeField(field.name, description, concept_type, default_value, has_default_value, field.init))

    return ConceptDictionaryType(
        set(getattr(annotation, "__is_a__", [])),
        annotation,
        (
            (docstring.short_description or "") + (docstring.long_description or "")
            if docstring.short_description and docstring.long_description
            else docstring.short_description or docstring.long_description or ""
        ),
        dict_fields,
        unset=unset,
    )



from typing import List, Optional

from kognitos.bdk.klang import KlangParser

from ...api import NounPhrase
from ...reflection import (BookCustomAuthenticationDescriptor,
                           CredentialDescriptor)


class BookCustomAuthenticationFactory:
    @classmethod
    def create(
        cls, id: str, noun_phrase_str: Optional[str], credentials: List[CredentialDescriptor], description: Optional[str], name: Optional[str]  # pylint: disable=redefined-builtin
    ) -> BookCustomAuthenticationDescriptor:
        if noun_phrase_str is None:
            noun_phrase = NounPhrase.from_snake_case(id)
        else:
            noun_phrases, _ = KlangParser.parse_noun_phrases(noun_phrase_str)
            noun_phrase = NounPhrase.from_tuple(noun_phrases[0])

        connect_name = noun_phrase.to_string() if name is None else name

        return BookCustomAuthenticationDescriptor(id=id, noun_phrase=noun_phrase, credentials=credentials, description=description, name=connect_name)



import inspect
from functools import wraps


def discover(fn):
    if not inspect.isfunction(fn):
        raise TypeError("The discover decorator can only be applied to functions.")

    fn.__discover__ = True
    return wraps(fn)(fn)



import inspect
from typing import List, Optional, Type

import uritemplate

from ..docstring import DocstringParser
from ..reflection import (BookOAuthAuthenticationDescriptor,
                          OauthArgumentDescriptor, OAuthFlow, OAuthProvider)


class oauth:  # pylint: disable=invalid-name
    id: str
    provider: OAuthProvider
    flows: List[OAuthFlow]
    authorize_endpoint: str
    token_endpoint: str
    scopes: Optional[List[str]]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        id: str,  # pylint: disable=redefined-builtin
        provider: OAuthProvider,
        authorize_endpoint: str,
        token_endpoint: str,
        flows: Optional[List[OAuthFlow]] = None,
        scopes: Optional[List[str]] = None,
    ):
        self.id = id
        self.provider = provider
        self.flows = flows if flows else [OAuthFlow.AUTHORIZATION_CODE]
        self.authorize_endpoint = authorize_endpoint
        self.token_endpoint = token_endpoint
        self.scopes = scopes

    def __call__(self, cls: Type):
        def decorator(cls):
            if not inspect.isclass(cls):
                raise TypeError("The oauth decorator can only be applied to classes.")

            if not hasattr(cls, "__oauth__"):
                cls.__oauth__ = []

            if cls.__doc__:
                parsed_docstring = DocstringParser.parse(cls.__doc__)
            else:
                parsed_docstring = None

            additional_arg_names = list(set(uritemplate.variables(self.authorize_endpoint)).union(set(uritemplate.variables(self.token_endpoint))))
            additional_args = []

            for additional_arg in additional_arg_names:
                parsed_param = parsed_docstring.param_by_name(additional_arg) if parsed_docstring else None

                additional_args.append(
                    OauthArgumentDescriptor(
                        id=parsed_param.name if parsed_param else additional_arg,  # type: ignore
                        name=(parsed_param.label or parsed_param.name) if parsed_param else additional_arg,  # type: ignore
                        description=parsed_param.description if parsed_param else None,
                    )
                )

            cls.__oauth__.append(
                BookOAuthAuthenticationDescriptor(
                    id=self.id,
                    description=None,
                    provider=self.provider,
                    flows=self.flows,
                    authorize_endpoint=self.authorize_endpoint,
                    token_endpoint=self.token_endpoint,
                    scopes=self.scopes,
                    name="oauth",
                    arguments=additional_args,
                )
            )

            return cls

        if cls is None:
            return decorator

        return decorator(cls)



from ..typing import Sensitive
from .discoverable import Discoverable
from .errors import NotFoundError, TypeMismatchError
from .filter import (FilterBinaryExpression, FilterBinaryOperator,
                     FilterExpression, FilterExpressionVisitor,
                     FilterUnaryExpression, FilterUnaryOperator,
                     NounPhrasesExpression, ValueExpression)
from .noun_phrase import NounPhrase, NounPhrases
from .questions import (Question, ask, clear_answers, get_from_context,
                        set_answer, unset_answer)

__all__ = [
    "NotFoundError",
    "TypeMismatchError",
    "FilterBinaryExpression",
    "FilterBinaryOperator",
    "FilterExpression",
    "FilterExpressionVisitor",
    "FilterUnaryExpression",
    "FilterUnaryOperator",
    "NounPhrase",
    "NounPhrases",
    "Sensitive",
    "NounPhrasesExpression",
    "ValueExpression",
    "ask",
    "clear_answers",
    "get_from_context",
    "set_answer",
    "unset_answer",
    "Question",
    "Discoverable",
]



from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from typing import List, Optional, Set, Type


class ConceptType(ABC, metaclass=ABCMeta):
    @property
    def name(self):
        return str(self)

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other) -> bool:
        pass

    @abstractmethod
    def __hash__(self) -> int:
        pass

    @abstractmethod
    def simplify(self) -> ConceptType:
        pass

    @abstractmethod
    def children(self) -> List[ConceptType]:
        pass

    @abstractmethod
    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        pass

    def find(self, target_type: Type[ConceptType], seen: Optional[Set[ConceptType]] = None) -> List[ConceptType]:
        if seen is None:
            seen = set()

        if self in seen:
            return []

        seen.add(self)
        instances = []

        for child in self.children():
            if isinstance(child, target_type):
                instances.append(child)

            instances.extend(child.find(target_type, seen))

        return list(set(instances))

    def flatten(self, **_kwargs) -> List[ConceptType]:
        return [self]



from dataclasses import dataclass
from typing import List, Optional, Set

from kognitos.bdk.klang import KlangParser
from kognitos.bdk.klang.parser_signature import ParserSignature

from ..api.noun_phrase import NounPhrase


@dataclass
class BookProcedureSignature:
    english: Optional[str]
    verbs: List[str]
    object: List[NounPhrase]
    preposition: Optional[str]
    outputs: Optional[List[List[NounPhrase]]]
    target: Optional[List[NounPhrase]]
    proper_nouns: Optional[Set[str]]

    @classmethod
    def from_parser_signature(cls, parser_signature: ParserSignature) -> "BookProcedureSignature":
        return BookProcedureSignature(
            english=parser_signature.english,
            verbs=parser_signature.verbs,
            object=[NounPhrase.from_tuple(t) for t in parser_signature.object],
            preposition=parser_signature.preposition,
            outputs=[[NounPhrase.from_tuple(t) for t in c] for c in parser_signature.outputs] if parser_signature.outputs else None,
            target=[NounPhrase.from_tuple(t) for t in parser_signature.target] if parser_signature.target else None,
            proper_nouns=parser_signature.proper_nouns,
        )

    @classmethod
    def from_english(cls, english: str) -> "BookProcedureSignature":
        parser_signature = KlangParser.parse_signature(english)

        return BookProcedureSignature.from_parser_signature(parser_signature)



from typing import List, Optional, Type

from .base import ConceptType


class ConceptTableTypeColumn:
    name: str
    description: Optional[str]
    t: ConceptType

    def __init__(self, name: str, description: Optional[str] = None, t: Optional[ConceptType] = None):
        self.name = name
        self.description = description
        self.t = t

    def __repr__(self) -> str:
        return f"{self.name}: {self.t!r}"

    def __str__(self) -> str:
        return f"{self.name}: {self.t}"

    def __eq__(self, other) -> bool:
        if self.name != other.name:
            return False
        if self.t != other.t:
            return False
        if self.description != other.description:
            return False
        return True

    def __hash__(self) -> int:
        return hash((self.name, self.description, self.t))


class ConceptTableType(ConceptType):
    description: Optional[str]
    columns: List[ConceptTableTypeColumn]
    concrete: Type

    def __init__(self, description: Optional[str], columns: List[ConceptTableTypeColumn], concrete: Type):
        self.description = description
        self.columns = columns
        self.concrete = concrete

    def simplify(self) -> ConceptType:
        simplified_columns = [ConceptTableTypeColumn(field.name, field.description, field.t.simplify()) for field in self.columns]

        return ConceptTableType(self.description, simplified_columns, self.concrete)

    def children(self) -> List[ConceptType]:
        return [column.t for column in self.columns]

    def __repr__(self) -> str:
        return f"Table[{', '.join([column.__repr__() for column in self.columns])}]"

    def __str__(self) -> str:
        return "table"

    def replace(self, old: ConceptType, new: ConceptType) -> ConceptType:
        if self == old:
            return new

        new_columns = [
            ConceptTableTypeColumn(column.name, column.description, new) if column.t == old else ConceptTableTypeColumn(column.name, column.description, old)
            for column in self.columns
        ]

        return ConceptTableType(self.description, new_columns, self.concrete)

    def __eq__(self, other) -> bool:
        if isinstance(other, ConceptTableType):
            return self.columns == other.columns
        return False

    def __hash__(self) -> int:
        combined_hash = hash("Table")
        for field in self.columns:
            combined_hash ^= hash(field)
        return combined_hash



from dataclasses import dataclass
from typing import Optional

from ..types import CredentialType


@dataclass
class CredentialDescriptor:
    id: str
    type: CredentialType
    label: Optional[str]
    description: Optional[str]



from typing import List, Optional

from docstring_parser import Docstring as BaseDocstring
from docstring_parser import DocstringMeta as BaseDocstringMeta
from docstring_parser import DocstringParam as BaseDocstringParam

from kognitos.bdk.klang import KlangParser

from ..api.noun_phrase import NounPhrase
from .example import DocstringExample


class DocstringConcept:
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, meta: Optional[BaseDocstringMeta]):
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None

    @property
    def noun_phrases(self) -> Optional[List[NounPhrase]]:
        try:
            if self._meta:
                noun_phrases, _ = KlangParser.parse_determiner_noun_phrases(self._meta.args[1])
                return [NounPhrase.from_tuple(np) for np in noun_phrases] if noun_phrases else None
        except SyntaxError:
            pass

        return None


class DocstringAttribute:
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, meta: Optional[BaseDocstringMeta]):
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None


class DocstringEnumMember:
    _name: str
    _member_meta: Optional[BaseDocstringMeta]
    _label_meta: Optional[BaseDocstringMeta]

    def __init__(self, member_meta: Optional[BaseDocstringMeta], label_meta: Optional[BaseDocstringMeta], name: Optional[str] = None):
        self._name = name
        self._member_meta = member_meta
        self._label_meta = label_meta

    @property
    def name(self) -> str:
        """
        Resolve the name of the enum member. It cascades through the possible data sources:

        1. The name of the enum member in the member meta
        2. The name of the enum member
        3. The name of the enum member in the label meta
        """
        if self._member_meta:
            return self._member_meta.args[1]

        if self._name:
            return self._name

        if self._label_meta:
            return self._label_meta.args[1]

        raise ValueError("Could not resolve name for enum member")

    @property
    def description(self) -> Optional[str]:
        if self._member_meta:
            return self._member_meta.description

        return None

    @property
    def noun_phrase(self) -> Optional[NounPhrase]:
        if self._label_meta:
            raw_noun_phrase_string = self._label_meta.description
            if raw_noun_phrase_string:
                noun_phrases, _ = KlangParser.parse_noun_phrases(raw_noun_phrase_string)
                return NounPhrase.from_tuple(noun_phrases[0])
            raise ValueError(f"No noun phrase found for enum member {self.name}. You probably have a malformed label docstring.")

        if self.name:
            raw_noun_phrase_string = self.name
            return NounPhrase.from_snake_case(raw_noun_phrase_string.lower())

        return None


class DocstringParam:
    _docstring: Optional[BaseDocstringParam]
    _meta: Optional[BaseDocstringMeta]

    def __init__(self, docstring: Optional[BaseDocstringParam], meta: Optional[BaseDocstringMeta]):
        self._docstring = docstring
        self._meta = meta

    @property
    def name(self) -> Optional[str]:
        if self._docstring:
            return self._docstring.arg_name

        if self._meta:
            return self._meta.args[1]

        return None

    @property
    def description(self) -> Optional[str]:
        if self._docstring:
            return self._docstring.description

        return None

    @property
    def label(self) -> Optional[str]:
        if self._meta:
            return self._meta.description

        return None


class Docstring:
    docstring: BaseDocstring

    def __init__(self, docstring: BaseDocstring):
        self.docstring = docstring

    @property
    def author(self) -> Optional[str]:
        for meta in self.docstring.meta:
            if "author" in meta.args:
                return meta.description
        return None

    @property
    def short_description(self) -> Optional[str]:
        return self.docstring.short_description

    @property
    def long_description(self) -> Optional[str]:
        return self.docstring.long_description

    @property
    def returns(self) -> Optional[str]:
        return self.docstring.returns.description if self.docstring.returns else None

    @property
    def input_concepts(self) -> List[DocstringConcept]:
        inputs = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "inputs":
                    inputs.append(DocstringConcept(meta))

        return inputs

    @property
    def examples(self) -> List[DocstringExample]:
        return [DocstringExample(example) for example in self.docstring.examples]

    @property
    def output_concepts(self) -> List[DocstringConcept]:
        outputs = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "outputs":
                    outputs.append(DocstringConcept(meta))

        return outputs

    @property
    def attributes(self) -> List[DocstringAttribute]:
        attributes = []

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "attribute":
                    attributes.append(DocstringAttribute(meta))

        return attributes

    @property
    def params(self) -> List[DocstringParam]:
        def convert_param(param: BaseDocstringParam):
            param_meta = None
            for meta in self.docstring.meta:
                if len(meta.args) > 1:
                    if meta.args[0] == "label" and meta.args[1] == param.arg_name:
                        param_meta = meta
                        break

            return DocstringParam(param, param_meta)

        params = list(map(convert_param, self.docstring.params))

        for meta in self.docstring.meta:
            if len(meta.args) > 1:
                if meta.args[0] == "label":
                    already = [param for param in params if param.name == meta.args[1]]
                    if not already:
                        params.append(DocstringParam(None, meta))

        return params

    @property
    def enum_members(self) -> List[DocstringEnumMember]:
        members = []

        enum_metas = sorted([meta for meta in self.docstring.meta if len(meta.args) > 1 and meta.args[0] == "enum"], key=lambda x: x.args[1])
        enum_label_metas = sorted([meta for meta in self.docstring.meta if len(meta.args) > 1 and meta.args[0] == "enum label"], key=lambda x: x.args[1])
        seen_enum_label_metas = set()

        # Make sure to initialize enum members with the correct label meta if it exists
        for enum_meta in enum_metas:
            enum_label_meta = next((meta for meta in enum_label_metas if meta.args[1] == enum_meta.args[1]), None)
            if enum_label_meta:
                seen_enum_label_metas.add(enum_label_meta)

            members.append(DocstringEnumMember(enum_meta, enum_label_meta))

        # Initialize enum members with the correct label meta if it exists but was not seen before, i.e the enum meta for the key description was not set.
        for elm in set(enum_label_metas) - seen_enum_label_metas:
            members.append(DocstringEnumMember(None, elm))

        return members

    def enum_member_by_name(self, name: str) -> DocstringEnumMember:
        """
        Given a name, return the description of the enum key.

        If no member is found, return a new DocstringEnumMember with the given name.
        """
        return next((member for member in self.enum_members if member.name == name), DocstringEnumMember(None, None, name))

    def param_by_name(self, param: str) -> Optional[DocstringParam]:
        for doc_param in self.params:
            if doc_param.name == param:
                return doc_param

        return None

    def param_description_by_name(self, param: str) -> Optional[str]:
        for doc_param in self.params:
            if doc_param.name == param:
                return doc_param.description

        return None

    def input_concept_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[DocstringConcept]:
        for doc_concept in self.input_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept

        return None

    def input_description_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[str]:
        for doc_concept in self.input_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept.description

        return None

    def output_concept_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[DocstringConcept]:
        for doc_concept in self.output_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept

        return None

    def output_description_by_noun_phrases(self, noun_phrases: List[NounPhrase]) -> Optional[str]:
        for doc_concept in self.output_concepts:
            if doc_concept.noun_phrases == noun_phrases:
                return doc_concept.description

        return None



from typing import Optional

from ..reflection.types import ConceptType


class TypeMismatchError(Exception):
    where: str
    expected: Optional[ConceptType]

    def __init__(self, where: str, expected: ConceptType):
        self.where = where
        self.expected = expected
        super().__init__(f"type mismatch on {where} expected {expected}")


class NotFoundError(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)
